open Unix
open Lazy
open Ast
open Format
open Filename
open Hammer

let make_formatter =
  let fd = openfile ("./ground.ml") [O_CREAT; O_TRUNC; O_WRONLY]
      0o755 in
  let oc = out_channel_of_descr fd in
  set_binary_mode_out oc false;
  formatter_of_out_channel oc 

let chop p =
  let variables =
    let rec f n =
      let m = string_of_int n in
      Wait ("(m_"^m^", l_"^m^")", lazy (f (n+1))) in
    f 0 in
  instantiate (hammer p) variables

let merge_all = function
  | [] -> SMap.empty
  | [m] -> m
  | l ->
    let rec aux k = function
      | []   -> k SMap.empty
      | [m]  -> k m
      | m::l -> aux (fun m' -> SMap.union (fun _ (s,r) (u,t) -> Some (s$u,r$t))
                        m m') l in
    aux (fun x -> x) l

let rec grind fmt ((m,l) : disp_prog) =
  let (pre_m, pre_l, l) = var_fold (fun (p_m, p_l, l') x ->
      match String.split_on_char ',' x with
      | [x] -> (p_m, p_l, x::l')
      | [m;l] -> ((String.sub m 1 (String.length m - 1))::p_m,
                  (String.sub l 1 (String.length l - 2))::p_l, l')
      | _ -> assert false ) ([], [""], []) l in
  let rec elim_last l' = function
    | [] | [_] -> l'
    | a::l -> elim_last (a::l') l in
  let pre_l = String.concat " $ " (if l=[] then elim_last [] pre_l else pre_l) in

  let grind_list call fmt l =
    let rec mince_list fmt = function
      | []   -> fprintf fmt ""
      | [a]  -> fprintf fmt "%a" call a
      | a::l -> fprintf fmt "%a; %a" call a mince_list l in
    fprintf fmt "@[[%a]@]" mince_list l in

  let mince_send fmt = function
    | DSend p -> fprintf fmt "Send %a" grind p in
  let mince_rec  fmt = function
    | DRec (a,p) -> fprintf fmt "Rec (fun %s -> %a)" a grind p in
  let mince_var litt fmt x =
    fprintf fmt "%s" (if litt then "\""^x^"\"" else x) in

  let prep_merge fmt next = match pre_m with
    | []  -> fprintf fmt "%a" next ()
    | [p] -> if SMap.is_empty m then fprintf fmt "%s" p
      else fprintf fmt "(merge_all [%a; %s])" next () p
    | _   -> if SMap.is_empty m then fprintf fmt "(merge_all %a)"
          (grind_list (mince_var false)) pre_m
      else fprintf fmt "(merge_all (%a::%a))" next ()
          (grind_list (mince_var false)) pre_m in

  let mince_channel fmt a (s,r) next = fun fmt ->
    fprintf fmt "@.(SMap.add \"%s\" @[(%a, @,%a)@] %a)" a
      (grind_list mince_send) s
      (grind_list mince_rec) r next in
  let grind_channel =
    channel_fold (mince_channel fmt) m
      (fun fmt () -> fprintf fmt "SMap.empty") in

  fprintf fmt "@[(@[%a, @]@[%s%a@])@]@?" prep_merge grind_channel pre_l 
    (if l = [] then (fun fmt _ -> fprintf fmt "%s"
                        (if pre_l = "" then "[]" else ""))
     else grind_list (mince_var true)) l
    
let grind_compile p =
  let fmt = make_formatter in
  let p = chop p in
  fprintf fmt "open Ast@.\
open Hammer@.\
open Format@.@.\
open Interpreter@.@.\
let ($) = List.rev_append@.@.\
let merge_all = function@.\
\  | [] -> SMap.empty@.\
\  | [m] -> m@.\
\  | l ->@.\
\    let rec aux k = function@.\
\      | []   -> k SMap.empty@.\
\      | [m]  -> k m@.\
\      | m::l -> aux (fun m' -> SMap.union (fun _ (s,r) (u,t) -> Some (s$u,r$t))@.\
\                        m m') l in@.\
\    aux (fun x -> x) l@.\
let (ground : prog) = %a@.@." grind p
